import csv
import os
from . import REPR_INDENT

__all__ = ["Writer"]


class Writer:
    """CSV writer"""

    def __init__(self, path, columns, delimiter=","):
        self.path = path
        self.path_exists = os.path.exists(path)
        self.delimiter = delimiter
        if self.path_exists:
            with open(path, "r") as f:
                header = next(csv.reader(f, delimiter=self.delimiter))
            if set(header) == set(columns):
                self.columns = header
                self.write_header = False
            else:
                raise ValueError(
                    f"Header in file is inconsistent with columns: header: {header}; columns {columns}"
                )
        else:
            self.columns = columns
            self.write_header = True

        self.file = None
        self.writer = None

    def __enter__(self):

        mode = "a+" if self.path_exists else "w"
        self.file = open(self.path, mode, buffering=1)
        self.writer = csv.DictWriter(
            f=self.file, fieldnames=self.columns, delimiter=self.delimiter
        )
        if self.write_header:
            self.writer.writeheader()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()

    def write_row(self, row_dict):
        self.writer.writerow(row_dict)

    def __call__(self, row_dict):
        self.write_row(row_dict)

    def __repr__(self):
        out = []
        out.append(self.__class__.__name__)
        out.append(" " * REPR_INDENT + f"filepath: {self.path}")
        out.append(" " * REPR_INDENT + f"columns: {self.columns}")
        return "\n".join(out)



import torch
import torch.nn as nn
import torch.optim as optim

from .buffer import BufferBase

__all__ = [
    "MODEL",
    "OPTIMIZER",
    "SCHEDULER",
    "DATALOADER",
    "BUFFER",
    "COMPONENTS",
    "map_str_to_types",
    "map_types_to_str",
]

MODEL = (
    nn.Module,
    nn.ModuleList,
    nn.ModuleDict,
    torch.jit.ScriptModule,
)
OPTIMIZER = (optim.Optimizer,)
SCHEDULER = (optim.lr_scheduler._LRScheduler,)
DATALOADER = torch.utils.data.DataLoader
BUFFER = BufferBase
COMPONENTS = (
    MODEL,
    OPTIMIZER,
    SCHEDULER,
    DATALOADER,
    BUFFER,
)

map_str_to_types = {
    "model": MODEL,
    "optimizer": OPTIMIZER,
    "scheduler": SCHEDULER,
    "dataloader": DATALOADER,
    "buffer": BUFFER,
}

map_types_to_str = {v: k for k, v in map_str_to_types.items()}



from setuptools import find_packages, setup

# load readme
with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="torchliter",
    version="0.0.6",
    author="Chenchao Zhao",
    author_email="chenchao.zhao@gmail.com",
    description="A lightweight training tool for pytorch projects.",
    packages=find_packages(exclude=["tests"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["torch", "numpy"],
    license="MIT",
)



import inspect
from functools import partial
from .base import EngineBase
from .buffer import ScalarSmoother, to_buffer

__all__ = ["Automated"]


class Automated(EngineBase):
    """Automated Engine
    Given a forward generator function, `from_forward` will return an Automated engine class

    For example
    ==================

    import torch
    import torch.nn as nn
    import torch.nn.functional as F


    def classification(engine, batch):
        # the first arg must be a place holder for engine class

        engine.train()
        x, y = batch
        lgs = engine.model(x)
        loss = F.cross_entropy(lgs, y)

        yield "loss", loss.item()
        # metrics will be registered as buffers

        acc = (lgs.max(-1).indices == y).float().mean()

        yield "acc", acc.item()


    eng = Automated.from_forward(classification)

    # attach other components such as model, optimizer, dataloader, etc.
    eng.attach(model=nn.Linear(2, 2))
    ...

    """

    def forward(self, batch, **kwargs):
        raise NotImplementedError("Method `forward` must be implemented.")

    @to_buffer("buffer_registry")
    def per_batch(self, batch, **kwargs):
        return self.forward(batch, **kwargs)

    def attach(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def from_forward(cls, func, smooth_window=50, **kwargs):
        assert inspect.isgeneratorfunction(
            func
        ), "The forward function must be a generator function with first arg being engine class placeholder."
        smooth_window = max(int(smooth_window), 1)
        buffer_names = _find_outputs(func)
        eng = cls()
        eng.forward = partial(func, eng)
        eng.attach(**{n: ScalarSmoother(smooth_window, **kwargs) for n in buffer_names})
        return eng


def _find_outputs(func):
    source = inspect.getsource(func)
    names = []
    for line in source.splitlines():
        line = line.strip()
        if line.startswith("yield"):
            line = line.replace("yield", "")
            names.append(line.split(",")[0].strip()[1:-1])
    return names
